import { FunctionComponent } from 'react';
import { IDBPDatabase, DBSchema } from 'idb/with-async-ittr';

/**
 * ** Expectation Maximization Algorithm (EM，期望最大化算法) **
 *
 * 1. How to calculate the centroids
 *
 * There are a Array of centroids, each centroid is a vector( [v[1], v[2], ... , v[n]] ).
 *
 * For a centroid, it vector is [w[1], w[2], ... , w[n]], in this group of vector：
 * w(i) = (v(1)(i) + v(2)(i) + ... + v(n)(i)) / n
 *
 * 2. How to measure two points are close or not
 *
 * Assuming there are two points:  A(x1, x2, x3, ... , xn) and B(y1, y2, y3, ... , yn).
 * cosine(A, B) = (x1 * y1 + x2 * y2 + ... + xn * yn) / (||A|| * ||B||)
 * ||A|| = sqrt(x1^2 + x2^2 + ... + xn^2), ||B|| = sqrt(y1^2 + y2^2 + ... + yn^2)
 *
 * A and B is all positive, so 0 <= cosine(A, B) <= 1
 * When cos close to 1, those two points are close, otherwise not.
 */
declare type Vector = number[];

declare enum PredictDatabaseStatus {
    empty = 0,
    collecting = 1,
    computing = 2,
    finished = 3
}
interface PredictHeightDB extends DBSchema {
    'dynamic-height-list': {
        value: {
            weight: Vector;
            height: number;
        };
        key: string;
    };
}
declare type ListenerCallbackType = (allList: Vector[], itemToHeightMap: Record<string, number>) => void;
declare class PredictDatabase {
    weight: Map<keyof Vector, number>;
    DBStatus: PredictDatabaseStatus;
    waitToPredictList: Vector[];
    listeners: ((allList: Vector[], itemToHeightMap: Record<string, number>) => void)[];
    db: IDBPDatabase<PredictHeightDB> | null;
    capacity: number;
    constructor(capacity: number);
    initDB(): Promise<void>;
    initWeight(weight: Map<keyof Vector, number>): void;
    restoreFromCache(): void;
    initWaitToPredictList(list: Vector[]): Promise<void>;
    clearAllData(): void;
    addListener(callback: ListenerCallbackType): void;
    removeListener(callback: ListenerCallbackType): void;
    isReadyToPredict(): Promise<boolean>;
    addSample(index: number, height: number): Promise<void>;
    predict(centroids: Vector[], centroidsHeight: number[]): number[];
}

interface ChildItemProps {
    index: number;
}
interface VirtualListProps {
    children: FunctionComponent<ChildItemProps>;
    itemCount: number;
    itemHeight: number;
    dividedAreaNum?: number;
    factors?: Vector[];
    useDynamicHeight?: boolean;
    version?: number;
    createWorker?: () => Worker;
}
interface ListObserverProps {
    indexList: number[];
    children: FunctionComponent<ChildItemProps>;
    dividedAreaNum: number;
    isObserving: boolean;
    heights: HeightItem[];
    db?: PredictDatabase;
    onItemHeightChange?: (index: number, height: number) => void;
    useDynamicHeight?: boolean;
}
interface ItemRendererProps {
    children: FunctionComponent<ChildItemProps>;
    db?: PredictDatabase;
    index: number;
    onItemHeightChange?: (index: number, height: number) => void;
}
declare enum UnsupportedBehavior {
    immediate = 0,
    timeout = 1
}
interface HeightItem {
    type: 'default' | 'predict' | 'real';
    value: number;
}

declare const VirtualList: (props: VirtualListProps) => JSX.Element;

declare function beginPredict(allList: Vector[], itemToHeightMap: Record<string, number>, itemCount: number, factors: Vector[], heights: HeightItem[]): HeightItem[];

export { ChildItemProps, HeightItem, ItemRendererProps, ListObserverProps, UnsupportedBehavior, VirtualListProps, beginPredict, VirtualList as default };
