// src/VirtualList.tsx
import { useCallback, useEffect as useEffect6, useRef as useRef5, useState as useState3 } from "react";
import { useMemo as useMemo2 } from "react";
import useWorker from "use-worker-async";

// src/hooks/useIdleCallback.tsx
import { useEffect, useRef } from "react";

// src/interface.ts
var UnsupportedBehavior = /* @__PURE__ */ ((UnsupportedBehavior2) => {
  UnsupportedBehavior2[UnsupportedBehavior2["immediate"] = 0] = "immediate";
  UnsupportedBehavior2[UnsupportedBehavior2["timeout"] = 1] = "timeout";
  return UnsupportedBehavior2;
})(UnsupportedBehavior || {});

// src/hooks/useIdleCallback.tsx
var DEFAULT_TIMEOUT = 1e3;
function useIdleCallback(callback, options, unsupportedBehavior = 1 /* timeout */) {
  const handleId = useRef(null);
  useEffect(() => {
    if (isSupported()) {
      handleId.current = requestIdleCallback(callback, options);
    } else if (unsupportedBehavior === 1 /* timeout */) {
      handleId.current = window.setTimeout(callback, (options == null ? void 0 : options.timeout) || DEFAULT_TIMEOUT);
    } else {
      callback();
    }
    return () => {
      if (!handleId.current) {
        return;
      }
      if (isSupported()) {
        cancelIdleCallback(handleId.current);
      } else if (unsupportedBehavior === 1 /* timeout */) {
        clearTimeout(handleId.current);
      }
    };
  }, [callback, options == null ? void 0 : options.timeout]);
}
function isSupported() {
  return typeof globalThis !== "undefined" && "requestIdleCallback" in globalThis && "cancelIdleCallback" in globalThis && typeof requestIdleCallback === "function" && typeof cancelIdleCallback === "function";
}

// src/hooks/useIntersection.tsx
import { useEffect as useEffect2, useState } from "react";
function useIntersection(ref, options, isActive) {
  const [intersectionObserverEntry, setIntersectionObserverEntry] = useState(null);
  useEffect2(
    () => {
      if (isActive && ref.current && typeof IntersectionObserver === "function") {
        let observer = new IntersectionObserver((entries) => {
          setIntersectionObserverEntry(entries[0]);
        }, options);
        observer.observe(ref.current);
        return () => {
          setIntersectionObserverEntry(null);
          observer.disconnect();
        };
      }
    },
    [
      ref.current,
      Array.isArray(options.threshold) ? options.threshold.join(",") : options.threshold,
      options.root,
      options.rootMargin,
      isActive
    ]
  );
  return intersectionObserverEntry;
}

// src/utils.ts
function groupArray(inputStream, groupNum) {
  const bucketNum = Math.min(Math.max(1, groupNum), inputStream.length);
  const bucketSizeArr = new Array(bucketNum).fill(0);
  const bucket = Array.from({ length: bucketNum }, () => []);
  inputStream.forEach((_, index) => {
    bucketSizeArr[index % bucketNum]++;
  });
  let bucketIndex = 0;
  inputStream.forEach((item) => {
    if (bucketSizeArr[bucketIndex] === 0) {
      bucketIndex++;
    }
    bucket[bucketIndex].push(item);
    bucketSizeArr[bucketIndex]--;
  });
  return bucket;
}

// src/predictHeight/EM.ts
function beginIteration(featureVectors = [], centroidNum, heightRecord) {
  let centroids = extractPartVectors(featureVectors, centroidNum);
  for (let i = 0; i < 40; i++) {
    centroids = calculateCentroids(featureVectors, centroids);
  }
  const centroidsHeight = [];
  let indexOfGroup = Array.from({ length: centroids.length }, () => []);
  for (let i = 0; i < featureVectors.length; i++) {
    const index = findNearestCentroidIndex(featureVectors[i], centroids);
    indexOfGroup[index].push(i);
  }
  for (let i = 0; i < indexOfGroup.length; i++) {
    const aGroup = indexOfGroup[i];
    let height = 0;
    for (let i2 = 0; i2 < aGroup.length; i2++) {
      const cacheKey = featureVectors[aGroup[i2]].join("-");
      height += heightRecord[cacheKey] || 0;
    }
    centroidsHeight[i] = height / aGroup.length;
  }
  return {
    centroids,
    centroidsHeight
  };
}
function calculateCentroids(featureVectors, centroids) {
  let groups = groupByCentroids(featureVectors, centroids);
  const newCentroids = [];
  for (let i = 0; i < groups.length; i++) {
    newCentroids.push(calculateCenterPoint(groups[i]));
  }
  return newCentroids;
}
function groupByCentroids(featureVectors, centroids) {
  let groups = Array.from({ length: centroids.length }, () => []);
  for (let i = 0; i < featureVectors.length; i++) {
    const index = findNearestCentroidIndex(featureVectors[i], centroids);
    groups[index].push(featureVectors[i]);
  }
  return groups.filter((i) => i.length !== 0);
}
function calculateCenterPoint(featureVectors) {
  if (featureVectors.length === 0) {
    return [];
  }
  let centerPoint = new Array(featureVectors[0].length).fill(0);
  for (let i = 0; i < featureVectors.length; i++) {
    for (let j = 0; j < featureVectors[i].length; j++) {
      centerPoint[j] += featureVectors[i][j];
    }
  }
  return centerPoint.map((i) => i / featureVectors.length);
}
function findNearestCentroidIndex(point, centroids) {
  let minIndex = -1;
  let minIndexCos = 0;
  for (let i = 0; i < centroids.length; i++) {
    const cos = calculateCosine(point, centroids[i]);
    if (cos > minIndexCos) {
      minIndex = i;
      minIndexCos = cos;
    }
  }
  return minIndex;
}
var sqrtCache = /* @__PURE__ */ new Map();
function calculateCosine(x, y) {
  const consOfTwoPoint = x.reduce((acc, cur, index) => acc + cur * y[index], 0) / (getSqrtOfVector(x) * getSqrtOfVector(y));
  return consOfTwoPoint;
}
function getSqrtOfVector(vector) {
  const cacheKey = vector.join(",");
  if (sqrtCache.has(cacheKey)) {
    return sqrtCache.get(cacheKey);
  }
  const ret = Math.sqrt(vector.reduce((acc, cur) => acc + cur * cur, 0));
  sqrtCache.set(cacheKey, ret);
  return ret;
}
function extractPartVectors(featureVectors = [], size) {
  size = Math.max(1, Math.min(featureVectors.length, size));
  const uniqueIndices = /* @__PURE__ */ new Set();
  while (uniqueIndices.size < size) {
    const index = Math.floor(Math.random() * featureVectors.length);
    uniqueIndices.add(index);
  }
  return [...uniqueIndices].map((index) => featureVectors[index]);
}

// src/predictHeight/db.ts
import { openDB } from "idb/with-async-ittr";
var PredictDatabase = class {
  constructor(capacity) {
    this.weight = /* @__PURE__ */ new Map();
    this.DBStatus = 0 /* empty */;
    this.waitToPredictList = [];
    this.listeners = [];
    this.db = null;
    this.capacity = 1e3;
    this.capacity = capacity || 1e3;
  }
  async initDB() {
    this.db = await openDB("dynamic-virtual-list", 1, {
      upgrade(db) {
        db.createObjectStore("dynamic-height-list", {
          keyPath: "id",
          autoIncrement: true
        });
      }
    });
  }
  initWeight(weight) {
    this.weight = weight;
  }
  restoreFromCache() {
  }
  async initWaitToPredictList(list) {
    this.waitToPredictList = list;
    if (!this.db) {
      await this.initDB();
    }
  }
  clearAllData() {
    var _a;
    (_a = this.db) == null ? void 0 : _a.clear("dynamic-height-list");
  }
  addListener(callback) {
    this.listeners.push(callback);
  }
  removeListener(callback) {
    this.listeners = this.listeners.filter((item) => item !== callback);
  }
  async isReadyToPredict() {
    var _a;
    const count = await ((_a = this.db) == null ? void 0 : _a.count("dynamic-height-list"));
    if (!count || count && count <= this.capacity) {
      return false;
    }
    return true;
  }
  async addSample(index, height) {
    var _a, _b;
    if (this.waitToPredictList.length === 0 || !this.waitToPredictList[index]) {
      return;
    }
    const item = this.waitToPredictList[index];
    if (!this.db) {
      await this.initDB();
    }
    (_a = this.db) == null ? void 0 : _a.put("dynamic-height-list", {
      weight: item,
      height
    });
    if (await this.isReadyToPredict()) {
      if (this.DBStatus === 3 /* finished */) {
        return;
      }
      this.DBStatus = 3 /* finished */;
      const source = await ((_b = this.db) == null ? void 0 : _b.getAll("dynamic-height-list")) || [];
      let allList = [];
      let itemToHeightMap = {};
      source.forEach((item2) => {
        allList.push(item2.weight);
        itemToHeightMap[item2.weight.join("-")] = item2.height;
      });
      this.listeners.forEach((listener) => {
        listener(allList, itemToHeightMap);
      });
    }
  }
  predict(centroids, centroidsHeight) {
    const heights = [];
    this.waitToPredictList.forEach((item) => {
      const nearestIndex = findNearestCentroidIndex(item, centroids);
      heights.push(centroidsHeight[nearestIndex]);
    });
    return heights;
  }
};

// src/hooks/useDBPredictFinished.ts
import { useEffect as useEffect3, useRef as useRef2 } from "react";
function useDBPredictFinished(db, callback) {
  const saveCallbackRef = useRef2();
  useEffect3(() => {
    saveCallbackRef.current = callback;
  });
  useEffect3(() => {
    function listener(vector, itemToHeightMap) {
      var _a;
      (_a = saveCallbackRef.current) == null ? void 0 : _a.call(saveCallbackRef, vector, itemToHeightMap);
    }
    db.addListener(listener);
    return () => {
      db.removeListener(listener);
    };
  }, [db]);
}

// src/hooks/useTrackingValue.ts
import { useEffect as useEffect4, useRef as useRef3 } from "react";
function useTrackingValue(value, onChange) {
  const tracked = useRef3(value);
  const oldValue = tracked.current;
  useEffect4(() => {
    if (value !== oldValue) {
      tracked.current = value;
      onChange(value);
    }
  }, [onChange, value, oldValue]);
}

// src/hooks/useList.ts
import { useEffect as useEffect5, useMemo, useRef as useRef4, useState as useState2 } from "react";
function useList(initialListFn, deps) {
  const listRef = useRef4(initialListFn());
  const [_, forceUpdate] = useState2(0);
  const saveUpdate = () => forceUpdate((count) => (count + 1) % 1e9);
  useEffect5(() => {
    listRef.current = initialListFn();
    saveUpdate();
  }, deps);
  const actions = useMemo(() => {
    return {
      set: (newList) => {
        listRef.current = newList;
        saveUpdate();
      },
      update(index, value) {
        listRef.current[index] = value;
        saveUpdate();
      }
    };
  }, []);
  return [listRef.current, actions];
}

// src/VirtualList.tsx
import { Fragment, jsx } from "react/jsx-runtime";
var VIRTUAL_LIST_VERSION_KEY = "react-dynamic-list-version-cache-key";
var VirtualList = (props) => {
  const {
    itemCount,
    dividedAreaNum = 10,
    factors = [],
    itemHeight: itemMinHeight,
    useDynamicHeight = false,
    version = 0,
    createWorker
  } = props;
  const db = useMemo2(() => {
    return new PredictDatabase(itemCount * 10);
  }, []);
  const groupList = useMemo2(() => {
    const arr = Array.from({ length: itemCount }, (_, index) => index);
    return groupArray(arr, dividedAreaNum);
  }, [itemCount, dividedAreaNum]);
  const [heights, actions] = useList(
    () => Array.from({ length: itemCount }, () => ({ type: "default", value: itemMinHeight })),
    [itemCount, itemMinHeight]
  );
  useEffect6(() => {
    let previousVersion = 0;
    if (!useDynamicHeight) {
      return;
    }
    try {
      previousVersion = JSON.parse(localStorage.getItem(VIRTUAL_LIST_VERSION_KEY) || "0");
      if (previousVersion !== version) {
        db.clearAllData();
      }
    } catch (e) {
      console.error(e);
    }
  }, [version, useDynamicHeight]);
  const { workerRunner } = useWorker(createWorker, { autoTerminate: true });
  useEffect6(() => {
    if (useDynamicHeight && createWorker && Array.isArray(factors) && factors.length === itemCount) {
      db.initWaitToPredictList(factors);
    }
  }, [factors]);
  useDBPredictFinished(db, async (allList, itemToHeightMap) => {
    if (!useDynamicHeight) {
      return;
    }
    if (!createWorker) {
      return;
    }
    console.log("predicting...");
    const t0 = performance.now();
    const data = await workerRunner(allList, itemToHeightMap, itemCount, factors, heights);
    const t1 = performance.now();
    console.log(`has predicted, cost ${((t1 - t0) / 1e3 / 60).toFixed(2)} min`);
    if (Array.isArray(data) && data.length === heights.length) {
      actions.set(data);
    }
  });
  function handleItemHeightChange(index, height) {
    actions.update(index, {
      type: "real",
      value: height
    });
  }
  return /* @__PURE__ */ jsx(Fragment, {
    children: groupList.map((item, index) => /* @__PURE__ */ jsx(ListObserver, {
      db,
      useDynamicHeight,
      dividedAreaNum,
      indexList: item,
      heights,
      isObserving: true,
      children: props.children,
      onItemHeightChange: handleItemHeightChange
    }, index))
  });
};
var ListObserver = (props) => {
  const { indexList, children, dividedAreaNum, isObserving, heights } = props;
  const ref = useRef5(null);
  const groupedList = useMemo2(() => {
    return groupArray(indexList, dividedAreaNum);
  }, [indexList, dividedAreaNum]);
  const intersectionObserverEntry = useIntersection(ref, { threshold: 0 }, isObserving, `${indexList[0]}-${indexList[indexList.length - 1]}`);
  let minHeight = indexList.reduce((prev, cur) => {
    var _a;
    return prev + (((_a = heights[cur]) == null ? void 0 : _a.value) || 0);
  }, 0);
  function handleItemHeightChange(index, height) {
    var _a;
    if (heights[index].value === height || !props.useDynamicHeight) {
      return;
    }
    (_a = props.onItemHeightChange) == null ? void 0 : _a.call(props, index, height);
  }
  return /* @__PURE__ */ jsx("div", {
    ref,
    role: "list",
    style: { minHeight },
    children: intersectionObserverEntry && intersectionObserverEntry.isIntersecting ? /* @__PURE__ */ jsx(Fragment, {
      children: groupedList.map((subGroupedList, index) => {
        if (subGroupedList.length === 1) {
          return /* @__PURE__ */ jsx(ItemRenderer, {
            db: props.db,
            onItemHeightChange: handleItemHeightChange,
            index: subGroupedList[0],
            children: props.children
          }, index);
        } else {
          return /* @__PURE__ */ jsx(ListObserver, {
            db: props.db,
            onItemHeightChange: props.onItemHeightChange,
            indexList: subGroupedList,
            children,
            useDynamicHeight: props.useDynamicHeight,
            dividedAreaNum,
            heights,
            isObserving: intersectionObserverEntry.isIntersecting
          }, index);
        }
      })
    }) : null
  });
};
var ItemRenderer = (props) => {
  const [height, setHeight] = useState3(0);
  const measuredRef = useCallback((node) => {
    if (node !== null) {
      setHeight(node.getBoundingClientRect().height);
    }
  }, []);
  useIdleCallback(() => {
    var _a;
    if (height > 0) {
      (_a = props.db) == null ? void 0 : _a.addSample(props.index, height);
    }
  }, void 0, 0 /* immediate */);
  useTrackingValue(height, (newHeight) => {
    var _a;
    (_a = props.onItemHeightChange) == null ? void 0 : _a.call(props, props.index, newHeight);
  });
  return /* @__PURE__ */ jsx("div", {
    role: "item",
    "data-height": height,
    ref: measuredRef,
    children: props.children({ index: props.index })
  });
};
var VirtualList_default = VirtualList;

// src/predictHeight/worker.ts
function beginPredict(allList, itemToHeightMap, itemCount, factors, heights) {
  const { centroids, centroidsHeight } = beginIteration(allList, itemCount, itemToHeightMap);
  const predictHeights = [];
  factors.forEach((item) => {
    const nearestIndex = findNearestCentroidIndex(item, centroids);
    predictHeights.push(centroidsHeight[nearestIndex]);
  });
  const newHeights = [...heights];
  predictHeights.forEach((predictHeight, index) => {
    if (newHeights[index].type !== "real") {
      newHeights[index] = {
        type: "predict",
        value: predictHeight
      };
    }
  });
  return newHeights;
}
export {
  UnsupportedBehavior,
  beginPredict,
  VirtualList_default as default
};
